from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Assessments = CINTable.Assessments
AssessmentID = Assessments.AssessmentID
AssessmentFactors = Assessments.AssessmentFactors
AssessmentActualStartDate = Assessments.AssessmentActualStartDate
LAchildID = Assessments.LAchildID
CINdetailsID = Assessments.CINdetailsID

AssessmentFactorsList = CINTable.AssessmentFactorsList
AssessmentFactor = AssessmentFactorsList.AssessmentFactor


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of '8898'
    code="8898",
    # replace Assessments with the value in the module column of the excel sheet corresponding to this rule .
    module=CINTable.Assessments,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message=" The assessment has more than one parental or child factors with the same code",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[AssessmentFactors],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_ass = data_container[Assessments]
    df_asslist = data_container[AssessmentFactorsList]
    df_ass.index.name = "ROW_ID"
    df_ass.reset_index(inplace=True)

    df_ass_merged = df_ass.merge(
        df_asslist[["LAchildID", "CINdetailsID", "AssessmentID", "AssessmentFactor"]],
        on=["LAchildID", "CINdetailsID", "AssessmentID"],
    )

    # If there is more than one <AssessmentFactors> (N00181) for an assessment recorded, then none of the values should appear more than once for an assessment.

    # .duplicated() returns True in all the locations where LAchildID-CINdetailsID-AssessmentID-AssessmentFactors combination is a duplicate
    condition = df_ass_merged.duplicated(
        [LAchildID, CINdetailsID, AssessmentID, AssessmentFactor],
        keep=False,
    )

    df_issues = df_ass_merged[condition].drop_duplicates().reset_index()

    df_issues["ERROR_ID"] = tuple(
        zip(
            df_issues[LAchildID],
            df_issues[AssessmentID],
            df_issues[AssessmentFactors],
        )
    )

    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you do not change the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_3(
        table=Assessments, columns=[AssessmentFactors], row_df=df_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_assessments = pd.DataFrame(
        [  # child1
            {
                LAchildID: "child1",
                CINdetailsID: "cinID1",
                AssessmentActualStartDate: "01/01/2000",
                AssessmentFactors: ("111", "111"),  # fail: duplicate Assessment Factor
                AssessmentID: "1",  # fail: duplicate Assessment Factor
            },
            {
                LAchildID: "child1",
                CINdetailsID: "cinID1",
                AssessmentActualStartDate: "02/01/2000",
                AssessmentFactors: ("111", "222"),
                AssessmentID: "2",  # pass: different Assessment Factor
            },
        ]
    )
    sample_factors = pd.DataFrame(
        [  # child1
            {
                LAchildID: "child1",
                CINdetailsID: "cinID1",
                AssessmentFactor: "111",  # fail: duplicate Assessment Factor
                AssessmentID: "1",
            },
            {
                LAchildID: "child1",
                CINdetailsID: "cinID1",
                AssessmentFactor: "111",  # fail: duplicate Assessment Factor
                AssessmentID: "1",
            },
            {
                LAchildID: "child1",
                CINdetailsID: "cinID1",
                AssessmentFactor: "111",  # pass: different Assessment Factor
                AssessmentID: "2",
            },
            {
                LAchildID: "child1",
                CINdetailsID: "cinID2",
                AssessmentFactor: "222",  # pass: different Assessment Factor
                AssessmentID: "2",
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_assessments[AssessmentActualStartDate] = pd.to_datetime(
        sample_assessments[AssessmentActualStartDate],
        format="%d/%m/%Y",
        errors="coerce",
    )

    # Run rule function passing in our sample data
    result = run_rule(
        validate,
        {
            Assessments: sample_assessments,
            AssessmentFactorsList: sample_factors,
        },
    )

    # Use .type3_issues to check for the result of .push_type3_issues() which you used above.
    issues_list = result.type3_issues
    # Issues list contains the objects pushed in their respective order. Since push_type3 was only used once, there will be one object in issues_list.
    assert len(issues_list) == 1

    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == Assessments

    issue_columns = issues.columns
    assert issue_columns == [AssessmentFactors]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 1 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 1
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "1",
                    ("111", "111"),
                ),
                "ROW_ID": [0],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8898' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8898"
    assert (
        result.definition.message
        == " The assessment has more than one parental or child factors with the same code"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Assessments = CINTable.Assessments
LAchildID = Assessments.LAchildID
AssessmentFactors = Assessments.AssessmentFactors
CINdetailsID = Assessments.CINdetailsID
AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate


# define characteristics of rule
@rule_definition(
    # write the rule code here
    code="8869",
    # replace Assessments with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.Assessments,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="The assessment factors code “21” cannot be used in conjunction with any other assessment factors.",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        AssessmentFactors,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df = data_container[Assessments].copy()

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.
    df.reset_index(inplace=True)

    # lOGIC
    # If <AssessmentFactors> (N00181) = “21”, this must be the only <AssessmentFactors> (N00181) present.
    df_orig = df.copy()

    df = df[(df[AssessmentFactors].astype(str) == "21")]

    #  Merge tables
    df = df.merge(
        df_orig,
        how="left",
        on=[LAchildID, CINdetailsID, AssessmentAuthorisationDate],
        suffixes=["", "_orig"],
    )

    # Values that aren't 21 are now errors (using guidelines from rule 8790)
    df = df[df["AssessmentFactors_orig"] != "21"]

    failing_indices = df.set_index("ROW_ID_orig").index

    rule_context.push_issue(
        table=Assessments, field=AssessmentFactors, row=failing_indices
    )


def test_validate():
    # 0      #1      #2      #3     #4      #5      #6      #7
    ids = ["1", "1", "2", "3", "3", "4", "4", "5", "6", "6", "6"]
    cinid = ["1", "1", "2", "3", "3", "4", "4", "5", "6", "6", "6"]
    assessmentfactors = [
        "21",
        "AIND",  # id1, cinid1 fail. same assessment period has factor 21
        "NONE",
        pd.NA,
        "MOTH",
        "21",
        "AAAA",  # id4, cinid4 fail. same assessment period has factor 21
        "AA",
        "21",
        "14",  # ignored. not the same auth date as preceding.
        "15",
    ]
    # non-date placeholders can be used since this rule doesn't require date-object properties.
    assessmentauthorisationdate = [
        "1",
        "1",
        "2",
        "3",
        "3",
        "4",
        "4",
        "5",
        "2021-02-09",
        "2020-12-15",
        "2020-12-15",
    ]

    fake_df = pd.DataFrame(
        {
            "LAchildID": ids,
            "CINdetailsID": cinid,
            "AssessmentFactors": assessmentfactors,
            "AssessmentAuthorisationDate": assessmentauthorisationdate,
        }
    )

    result = run_rule(validate, {Assessments: fake_df})

    issues = list(result.issues)

    assert len(issues) == 2

    assert issues == [
        IssueLocator(CINTable.Assessments, AssessmentFactors, 1),
        IssueLocator(CINTable.Assessments, AssessmentFactors, 6),
    ]

    assert result.definition.code == "8869"
    assert (
        result.definition.message
        == "The assessment factors code “21” cannot be used in conjunction with any other assessment factors."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

CINdetails = CINTable.CINdetails
LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID
CINreferralDate = CINdetails.CINreferralDate
CINclosureDate = CINdetails.CINclosureDate
ReferralNFA = CINdetails.ReferralNFA

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    # write the rule code here
    code="8820",
    # replace CINdetails with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.CINdetails,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="The dates on the CIN episodes for this child overlap",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        CINreferralDate,
        CINclosureDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    # Replace CINdetails with the name of the table you need.
    df_cin = data_container[CINdetails]
    df_cin2 = data_container[CINdetails]

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_cin.index.name = "ROW_ID"
    df_cin2.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.
    df_cin.reset_index(inplace=True)
    df_cin2.reset_index(inplace=True)

    # ReferenceDate exists in the header table so we get header table too.
    df_ref = data_container[Header]
    ref_date_series = df_ref[ReferenceDate]

    # the make_census_period function generates the start and end date so that you don't have to do it each time.
    collection_start, reference_date = make_census_period(ref_date_series)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # If more than one <CINdetails> module is present for the same child, then for each module the <CINreferralDate> (N00100) must not fall;
    # a) between the <CINreferralDate> (N00100) and the <CINclosureDate> (N00102) if present; and
    # b) between the <CINreferralDate> (N00100) and the <ReferenceDate> (N00603) if the <CINclosureDate> is not present and <ReferralNFA> (N00112) = false or 0
    #
    # Note: the effect of this rule is that there cannot be overlapping referrals, although the end date of one referral may be the same as the start date of the following referral.

    #  Create dataframes which only have rows with CINreferralDate, and which should have one plan per row.
    df_cin = df_cin[df_cin[CINreferralDate].notna()]
    df_cin2 = df_cin2[df_cin2[CINreferralDate].notna()]

    #  Merge tables to test for overlaps
    df_merged = df_cin.merge(
        df_cin2,
        on=["LAchildID"],
        how="left",
        suffixes=("_cin", "_cin2"),
    )

    # Exclude rows where the ROW_ID is the same on both sides
    df_merged = df_merged[(df_merged["ROW_ID_cin"] != df_merged["ROW_ID_cin2"])]

    # Determine overlaps
    cin_started_after_start = (
        df_merged["CINreferralDate_cin"] >= df_merged["CINreferralDate_cin2"]
    )
    cin_started_before_end = (
        df_merged["CINreferralDate_cin"] < df_merged["CINclosureDate_cin2"]
    ) & df_merged["CINclosureDate_cin2"].notna()

    cin_started_before_refdate = (
        (df_merged["CINreferralDate_cin"] < reference_date)
        & df_merged["CINclosureDate_cin2"].isna()
        & df_merged["ReferralNFA_cin2"].str.lower().isin(["false", "0"])
    )

    df_merged = df_merged[
        cin_started_after_start & (cin_started_before_end | cin_started_before_refdate)
    ].reset_index()

    # create an identifier for each error instance.
    # In this case, the rule is checked for each CPPstartDate, in each CPplanDates group (differentiated by CP dates), in each child (differentiated by LAchildID)
    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged[LAchildID],
            df_merged["CINdetailsID_cin"],
            df_merged["CINdetailsID_cin2"],
        )
    )

    # The merges were done on copies of cpp_df so that the column names in dataframes themselves aren't affected by the suffixes.
    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_cin_issues = (
        df_cin.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin2_issues = (
        df_cin2.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cin2")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_3(
        table=CINdetails, columns=[CINreferralDate], row_df=df_cin_issues
    )
    rule_context.push_type_3(
        table=CINdetails,
        columns=[CINreferralDate, CINclosureDate],
        row_df=df_cin2_issues,
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2001"}]  # the census start date here will be 01/04/2000
    )

    sample_cin = pd.DataFrame(
        [  # child1
            {
                "LAchildID": "child1",
                "CINreferralDate": "26/05/2000",  # 0 Pass: not between "26/08/2000" and "31/03/2001"
                "CINclosureDate": "26/10/2000",
                "CINdetailsID": "cinID1",
                "ReferralNFA": "true",
            },
            {
                "LAchildID": "child1",
                "CINreferralDate": "26/08/2000",  # 1 Fail: between "26/05/2000" and "26/10/2000"
                "CINclosureDate": pd.NA,
                "CINdetailsID": "cinID12",
                "ReferralNFA": "0",
            },
            # child2
            {
                "LAchildID": "child2",
                "CINreferralDate": "26/05/2000",  # 2 pass, not between
                "CINclosureDate": "25/10/2000",
                "CINdetailsID": "cinID2",
                "ReferralNFA": "true",
            },
            {
                "LAchildID": "child2",
                "CINreferralDate": "26/10/2000",  # 3 pass, not between
                "CINclosureDate": "26/12/2000",
                "CINdetailsID": "cinID22",
                "ReferralNFA": "true",
            },
            # child3
            {
                "LAchildID": "child3",
                "CINreferralDate": "26/05/2000",  # 4 Pass: not between "26/08/2000" and "26/10/2000"
                "CINclosureDate": "26/10/2001",
                "CINdetailsID": "cinID3",
                "ReferralNFA": "true",
            },
            {
                "LAchildID": "child3",  # 5 Fail: between "26/05/2000" and "26/10/2001"
                "CINreferralDate": "26/08/2000",
                "CINclosureDate": "26/10/2000",
                "CINdetailsID": "cinID32",
                "ReferralNFA": "true",
            },
            # child4
            {
                "LAchildID": "child4",
                "CINreferralDate": "26/10/2000",  # 6 Ignore: between "26/09/2000" and ReferenceDate but ReferralNFA is true
                "CINclosureDate": "31/03/2001",
                "CINdetailsID": "cinID4",
                "ReferralNFA": "true",
            },
            {
                "LAchildID": "child4",
                "CINreferralDate": "26/09/2000",  # 7 Pass: not between "26/10/2000" and "31/03/2001"
                "CINclosureDate": pd.NA,
                "CINdetailsID": "cinID42",
                "ReferralNFA": "true",
            },
            # child 5
            {
                "LAchildID": "child5",
                "CINreferralDate": "08/07/2000",
                "CINclosureDate": "23/08/2000",
                "CINdetailsID": "cinID4",
                "ReferralNFA": "false",
            },
            {
                "LAchildID": "child5",
                "CINreferralDate": "05/05/2000",
                "CINclosureDate": "08/07/2000",
                "CINdetailsID": "cinID5",
                "ReferralNFA": "false",
            },
            # child 6 - to account for duplicated entries as per issue 372
            {
                "LAchildID": "child6",  # 10, fail, duplicated
                "CINreferralDate": "05/05/2000",
                "CINclosureDate": "08/07/2000",
                "CINdetailsID": "cinID5",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child6",  # 11, fail, duplicated
                "CINreferralDate": "05/05/2000",
                "CINclosureDate": "08/07/2000",
                "CINdetailsID": "cinID5",
                "ReferralNFA": "0",
            },
        ]
    )

    # If rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_cin[CINreferralDate] = pd.to_datetime(
        sample_cin[CINreferralDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin[CINclosureDate] = pd.to_datetime(
        sample_cin[CINclosureDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_header[ReferenceDate] = pd.to_datetime(
        sample_header[ReferenceDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            CINdetails: sample_cin,
            Header: sample_header,
        },
    )

    issues_list = result.type3_issues
    assert len(issues_list) == 2
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    issues = issues_list[0]

    # get table name and check it. Replace CINdetails with the name of your table.
    issue_table = issues.table
    assert issue_table == CINdetails

    # check that the right columns were returned. Replace CINreferralDate  with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [CINreferralDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 3

    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "cinID12",
                    "cinID1",
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child3",
                    "cinID32",
                    "cinID3",
                ),
                "ROW_ID": [5],
            },
            {
                "ERROR_ID": (
                    "child6",
                    "cinID5",
                    "cinID5",
                ),
                "ROW_ID": [10, 11],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8820' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8820"
    assert (
        result.definition.message
        == "The dates on the CIN episodes for this child overlap"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

CINdetails = CINTable.CINdetails
LAchildID = CINdetails.LAchildID
CINreferralDate = CINdetails.CINreferralDate
ReferralSource = CINdetails.ReferralSource


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of '8866'
    code="8866",
    # replace CINdetails with the value in the module column of the excel sheet corresponding to this rule .
    module=CINTable.CINdetails,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Source of Referral is missing or an invalid code",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        CINreferralDate,
        ReferralSource,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # Replace CINdetails with the name of the table you need.
    df = data_container[CINdetails]
    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df.index.name = "ROW_ID"
    df.reset_index(inplace=True)

    valid_referrals = [
        "1A",
        "1B",
        "1C",
        "1D",
        "2A",
        "2B",
        "3A",
        "3B",
        "3C",
        "3D",
        "3E",
        "3F",
        "4",
        "5A",
        "5B",
        "5C",
        "5D",
        "6",
        "7",
        "8",
        "9",
        "10",
    ]

    # If <CinReferralDate> (N00100) is on or after 1 April 2013 then <ReferralSource> (N00152) must be present and must be a valid code
    condition = (
        df[CINreferralDate] >= pd.to_datetime("01/04/2013", format="%d/%m/%Y")
    ) & ((df[ReferralSource].isna()) | (~df[ReferralSource].isin(valid_referrals)))
    # get all the data that fits the failing condition. Reset the index so that ROW_ID now becomes a column of df
    df_issues = df[condition].reset_index()

    # SUBMIT ERRORS
    # Generate a unique ID for each instance of an error.

    # Replace CPPstartDate and CPPendDate below with the columns concerned in your rule.
    link_id = tuple(zip(df_issues[LAchildID], df_issues[CINreferralDate]))
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    # Ensure that you do not change the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_1(
        table=CINdetails, columns=[CINreferralDate, ReferralSource], row_df=df_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.

    #  Fails rows 0, 1, and 3
    sample_cin = pd.DataFrame(
        [
            {
                "LAchildID": "ID1",  # Fail
                "CINreferralDate": "01/01/2020",
                "ReferralSource": "1Z",
            },
            {
                "LAchildID": "ID2",  # Fail
                "CINreferralDate": "01/01/2020",
                "ReferralSource": pd.NA,
            },
            {
                "LAchildID": "ID3",
                "CINreferralDate": "01/01/2000",
                "ReferralSource": "1A",
            },
            {
                "LAchildID": "ID4",
                "CINreferralDate": "01/01/2012",  # ignore
                "ReferralSource": "pd.NA",
            },
            {
                "LAchildID": "ID4",
                "CINreferralDate": "01/01/2020",
                "ReferralSource": "1C",
            },
        ]
    )
    sample_cin["CINreferralDate"] = pd.to_datetime(
        sample_cin["CINreferralDate"], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {CINdetails: sample_cin})

    # The result contains a NamedTuple of issues encountered
    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [CINreferralDate, ReferralSource]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df

    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "ID1",
                    pd.to_datetime("01/01/2020", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "ID2",
                    pd.to_datetime("01/01/2020", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    assert result.definition.code == "8866"
    assert (
        result.definition.message == "Source of Referral is missing or an invalid code"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

ChildProtectionPlans = CINTable.ChildProtectionPlans
CPPendDate = ChildProtectionPlans.CPPendDate
LAchildID = ChildProtectionPlans.LAchildID

ChildIdentifiers = CINTable.ChildIdentifiers
PersonDeathDate = ChildIdentifiers.PersonDeathDate
LAchildID = ChildIdentifiers.LAchildID


# define characteristics of rule
@rule_definition(
    code="8920",
    module=CINTable.ChildProtectionPlans,
    message="Child Protection Plan cannot end after the child’s Date of Death",
    affected_fields=[
        PersonDeathDate,
        CPPendDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df_cpp = data_container[ChildProtectionPlans].copy()
    df_child = data_container[ChildIdentifiers].copy()

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_cpp.index.name = "ROW_ID"
    df_child.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.
    df_cpp.reset_index(inplace=True)
    df_child.reset_index(inplace=True)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # If <PersonDeathDate> (N00108) is present, then <CPPendDate> (N00115) must be on or before <PersonDeathDate> (N00108)
    # Issues dfs should return rows where PersonDeathDate is less than or equal to the CPPendDate

    #  Create dataframes which only have rows with CP plans, and which should have one plan per row.
    df_cpp = df_cpp[df_cpp[CPPendDate].notna()]
    df_child = df_child[df_child[PersonDeathDate].notna()]

    #  Merge tables to get corresponding CP plan group for the child
    df_merged = df_cpp.merge(
        df_child,
        left_on=["LAchildID"],
        right_on=["LAchildID"],
        how="left",
        suffixes=("_cpp", "_child"),
    )

    #  Get rows where CPPendDate is after PersonDeathDate
    condition = df_merged[CPPendDate] > df_merged[PersonDeathDate]
    df_merged = df_merged[condition].reset_index()

    # create an identifier for each error instance.
    # In this case, the rule is checked for each CPPendDate against the PersonDeathDate for that child.
    # A child may have multiple CP Plans but only 1 should be current at anytime and requires a CPPendDate

    df_merged["ERROR_ID"] = tuple(
        zip(df_merged[LAchildID], df_merged[CPPendDate], df_merged[PersonDeathDate])
    )

    # The merges were done on copies of df_cpp and df_child so that the column names in dataframes themselves aren't affected by the suffixes.
    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_cpp_issues = (
        df_cpp.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cpp")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_reviews_issues = (
        df_child.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_child")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_2(
        table=ChildProtectionPlans, columns=[CPPendDate], row_df=df_cpp_issues
    )
    rule_context.push_type_2(
        table=ChildIdentifiers, columns=[PersonDeathDate], row_df=df_reviews_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_cpp = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CPPendDate": "26/05/2000",  # Fails as PersonDeathDate is before the CPPendDate
            },
            {
                "LAchildID": "child2",
                "CPPendDate": "27/06/2002",  # Fails as PersonDeathDate is before the CPPendDate
            },
            {
                "LAchildID": "child3",
                "CPPendDate": "07/02/2001",  # Fails as PersonDeathDate is before the CPPendDate
            },
            {
                "LAchildID": "child4",
                "CPPendDate": "26/05/2000",  # Passes as PersonDeathDate is after the CPPendDate
            },
            {
                "LAchildID": "child5",
                "CPPendDate": "26/05/2000",  # Passes as PersonDeathDate is after the CPPendDate
            },
            {
                "LAchildID": "child6",
                "CPPendDate": pd.NA,  # Ignored as rows with no CPPendDate are dropped (this is picked up by other rules)
            },
            {
                "LAchildID": "child7",
                "CPPendDate": "14/03/2001",  # Ignored as rows with no PersonDeathDate are dropped (this is picked up by other rules)
            },
        ]
    )
    sample_children = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "PersonDeathDate": "25/05/2000",  # Fails as PersonDeathDate is before the CPPendDate
            },
            {
                "LAchildID": "child2",
                "PersonDeathDate": "29/05/2000",  # Fails as PersonDeathDate is before the CPPendDate
            },
            {
                "LAchildID": "child3",
                "PersonDeathDate": "26/03/2000",  # Fails as PersonDeathDate is before the CPPendDate
            },
            {
                "LAchildID": "child4",
                "PersonDeathDate": "30/05/2000",  # Passes as PersonDeathDate is after the CPPendDate
            },
            {
                "LAchildID": "child5",
                "PersonDeathDate": "27/05/2000",  # Passes as PersonDeathDate is after the CPPendDate
            },
            {
                "LAchildID": "child6",
                "PersonDeathDate": "26/05/2000",  # Ignored as rows with no CPPendDate are dropped (this is picked up by other rules)
            },
            {
                "LAchildID": "child7",
                "PersonDeathDate": pd.NA,  # Ignored as rows with no PersonDeathDate are dropped (this is picked up by other rules)
            },
        ]
    )

    # If rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_cpp[CPPendDate] = pd.to_datetime(
        sample_cpp[CPPendDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_children["PersonDeathDate"] = pd.to_datetime(
        sample_children["PersonDeathDate"], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            ChildProtectionPlans: sample_cpp,
            ChildIdentifiers: sample_children,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type2_issues
    assert len(issues_list) == 2
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values. the tuple in location 1 will contain the Reviews columns because that's the second thing pushed above.
    issues = issues_list[1]

    # get table name and check it. Replace Reviews with the name of your table.
    issue_table = issues.table
    assert issue_table == ChildIdentifiers

    # check that the right columns were returned. Replace CPPreviewDate  with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [PersonDeathDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 3 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 3
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",  # ChildID
                    # CPP End Date
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                    # Person Death Date
                    pd.to_datetime("25/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child2",  # ChildID
                    # CPP End Date
                    pd.to_datetime("27/06/2002", format="%d/%m/%Y", errors="coerce"),
                    # Person Death Date
                    pd.to_datetime("29/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child3",  # ChildID
                    # CPP End Date
                    pd.to_datetime("07/02/2001", format="%d/%m/%Y", errors="coerce"),
                    # Person Death Date
                    pd.to_datetime("26/03/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [2],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 2885 with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8920"
    assert (
        result.definition.message
        == "Child Protection Plan cannot end after the child’s Date of Death"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

CINplanDates = CINTable.CINplanDates
CINPlanStartDate = CINplanDates.CINPlanStartDate
LAchildID = CINplanDates.LAchildID

CINdetails = CINTable.CINdetails
CINreferralDate = CINdetails.CINreferralDate


@rule_definition(
    code="4015",
    module=CINTable.CINplanDates,
    message="The CIN Plan start date cannot be before the referral date",
    affected_fields=[
        CINPlanStartDate,
        CINreferralDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    """
    Where present, the <CINPlanStartDate> (N00689) must be on or after the <CINReferralDate> (N00100)
    """
    df_cindetail = data_container[CINdetails].copy()
    df_cinplan = data_container[CINplanDates].copy()

    df_cindetail.index.name = "ROW_ID"
    df_cinplan.index.name = "ROW_ID"

    df_cindetail.reset_index(inplace=True)
    df_cinplan.reset_index(inplace=True)

    # Where present, the <CINPlanStartDate> (N00689) must be on or after the <CINReferralDate> (N00100)
    df_cinplan = df_cinplan[df_cinplan[CINPlanStartDate].notna()]

    df_merged = df_cindetail.merge(
        df_cinplan,
        how="inner",
        on=["LAchildID", "CINdetailsID"],
        suffixes=["_det", "_plan"],
    )

    df_merged.query(r"CINPlanStartDate < CINreferralDate", inplace=True)

    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged[LAchildID],
            df_merged[CINPlanStartDate],
            df_merged[CINreferralDate],
        )
    )

    df_cindet_issues = (
        df_cinplan.merge(
            df_merged, how="inner", left_on="ROW_ID", right_on="ROW_ID_plan"
        )
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_cinplan_issues = (
        df_cindetail.merge(
            df_merged, how="inner", left_on="ROW_ID", right_on="ROW_ID_det"
        )
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=CINplanDates, columns=[CINPlanStartDate], row_df=df_cindet_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[CINreferralDate], row_df=df_cinplan_issues
    )


def test_validate():
    plan_is = (
        # ID     #CINID   #PlanStartDate
        ("1", "45", "2020-05-05"),  # 0
        ("4", "55", "2019-04-20"),  # 1
        ("67", "66", "2014-03-21"),  # 2
        ("69", "67", "2018-04-20"),  # 3
        ("69", "67", pd.NA),  # 4
        ("167", "166", "2014-03-21"),  # 5
    )

    cin_is = (
        # ID     #CINID   #CIN Ref Date
        ("1", "44", "2017-05-05"),  # A  0
        ("4", "55", "2019-04-20"),  # B  1
        ("67", "66", "2016-03-21"),  # C  2
        ("67", "67", "2015-03-21"),  # D  3
        ("69", "67", "2018-04-20"),  # E  4
        ("70", "69", "2015-04-20"),  # F  5
        ("167", "166", "2015-02-21"),  # G  6
    )

    fake_cinplan = pd.DataFrame(
        {
            "LAchildID": [x[0] for x in plan_is],
            "CINdetailsID": [x[1] for x in plan_is],
            "CINPlanStartDate": [x[2] for x in plan_is],
        }
    )
    fake_cindetail = pd.DataFrame(
        {
            "LAchildID": [x[0] for x in cin_is],
            "CINdetailsID": [x[1] for x in cin_is],
            "CINreferralDate": [x[2] for x in cin_is],
        }
    )
    fake_cinplan[CINPlanStartDate] = pd.to_datetime(
        fake_cinplan[CINPlanStartDate], format=r"%Y-%m-%d", errors="coerce"
    )
    fake_cindetail["CINreferralDate"] = pd.to_datetime(
        fake_cindetail["CINreferralDate"], format=r"%Y-%m-%d", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            CINplanDates: fake_cinplan,
            CINdetails: fake_cindetail,
        },
    )

    issues_list = result.type2_issues

    assert len(issues_list) == 2
    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [CINreferralDate]

    issue_rows = issues.row_df

    assert len(issue_rows) == 2

    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "67",  # ChildID
                    # Start Date
                    pd.to_datetime("21/03/2014", format=r"%d/%m/%Y", errors="coerce"),
                    # Ref date
                    pd.to_datetime("21/03/2016", format=r"%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "167",  # ChildID
                    # Start date
                    pd.to_datetime("21/03/2014", format=r"%d/%m/%Y", errors="coerce"),
                    # Ref date
                    pd.to_datetime("21/02/2015", format=r"%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [6],
            },
        ]
    )
    issue_rows.sort_values(["ROW_ID"], ignore_index=True, inplace=True)
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "4015"
    assert (
        result.definition.message
        == "The CIN Plan start date cannot be before the referral date"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import england_working_days, make_census_period

CINdetails = CINTable.CINdetails

ReferralNFA = CINdetails.ReferralNFA
LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID
CINreferralDate = CINdetails.CINreferralDate

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


@rule_definition(
    code="8569Q",
    module=CINTable.CINdetails,
    rule_type=RuleType.QUERY,
    message="A case with referral date before one working day prior to the collection start date must not be flagged as a no further action case",
    affected_fields=[
        ReferralNFA,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_cin = data_container[CINdetails].copy()
    df_cin.index.name = "ROW_ID"

    df_cin.reset_index(inplace=True)

    header = data_container[Header]
    ref_date_series = header[ReferenceDate]
    collection_start, collection_end = make_census_period(ref_date_series)

    # If <CINreferralDate> (N00100) is before [Start_of_Census_Year] minus 1 working day, <ReferralNFA> (N00112) must be false
    df_cin_issues = df_cin[
        df_cin[CINreferralDate] < (collection_start - england_working_days(1))
    ]

    df_cin_issues = df_cin_issues[
        ~df_cin_issues[ReferralNFA].isin(["false", "0"])
    ].reset_index()

    df_cin_issues["ERROR_ID"] = tuple(
        zip(
            df_cin_issues[LAchildID],
            df_cin_issues[CINdetailsID],
            df_cin_issues[CINreferralDate],
        )
    )

    df_issues = (
        df_cin_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=CINdetails, columns=[CINreferralDate], row_df=df_issues
    )


def test_validate():
    sample_cin_details = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINreferralDate": "26/10/1880",
                "ReferralNFA": "true",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child2",
                "CINreferralDate": "26/10/2001",
                "ReferralNFA": "true",
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child3",
                "CINreferralDate": "26/10/1880",
                "ReferralNFA": "false",
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child4",
                "CINreferralDate": "26/10/1999",
                "ReferralNFA": "false",
                "CINdetailsID": "cinID4",
            },
            {
                "LAchildID": "child5",
                "CINreferralDate": pd.NA,
                "ReferralNFA": "false",
                "CINdetailsID": "cinID5",
            },
        ]
    )
    sample_cin_details["CINreferralDate"] = pd.to_datetime(
        sample_cin_details["CINreferralDate"], format="%d/%m/%Y", errors="coerce"
    )
    sample_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2000"}]  # the census start date here will be 01/04/2000
    )

    result = run_rule(
        validate,
        {
            CINdetails: sample_cin_details,
            Header: sample_header,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 1
    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [CINreferralDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 1
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",  # ChildID
                    "cinID1",  # CINdetailsID
                    # corresponding CPPstartDate
                    pd.to_datetime("26/10/1880", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8569Q"
    assert (
        result.definition.message
        == "A case with referral date before one working day prior to the collection start date must not be flagged as a no further action case"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py
# Replace ChildProtectionPlans with the table name, and LAChildID with the column name you want.

ChildProtectionPlans = CINTable.ChildProtectionPlans
LAchildID = ChildProtectionPlans.LAchildID
CPPstartDate = ChildProtectionPlans.CPPstartDate
CPPendDate = ChildProtectionPlans.CPPendDate


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of '8840'
    code="8840",
    # replace ChildProtectionPlans with the value in the module column of the excel sheet corresponding to this rule .
    module=CINTable.ChildProtectionPlans,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Child Protection Plan cannot start and end on the same day",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[ChildProtectionPlans, CPPstartDate, CPPendDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # Replace ChildProtectionPlans with the name of the table you need.
    df = data_container[ChildProtectionPlans]
    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df.index.name = "ROW_ID"

    # LOGIC
    # Within a <ChildProtectionPlans> group, <CPPstartDate> (N00105) must not be the same as the <CPPendDate> (N00115)

    #  Determine if the dates are the same by finding if the difference between dates is 0
    condition = df["CPPstartDate"] == df["CPPendDate"]
    # get all the data that fits the failing condition. Reset the index so that ROW_ID now becomes a column of df
    df_issues = df[condition].reset_index()

    # SUBMIT ERRORS
    # Generate a unique ID for each instance of an error. In this case,
    # - If only LAchildID is used as an identifier, multiple instances of the error on a child will be understood as 1 instance.
    # We don't want that because in reality, a child can have multiple instances of an error.
    # - If we use the LAchildID-CPPstartDate combination, that artificially cancels out the instances where a start date repeats for the same child.
    # Another rule checks for that condition. Not this one.
    # - It is very unlikely that a combination of LAchildID-CPPstartDate-CPPendDate will repeat in the DataFrame.
    # Hence, it can be used as a unique identifier of the row.

    # Replace CPPstartDate and CPPendDate below with the columns concerned in your rule.
    link_id = tuple(
        zip(df_issues[LAchildID], df_issues[CPPstartDate], df_issues[CPPendDate])
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    # Ensure that you do not change the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_1(
        table=ChildProtectionPlans, columns=[CPPstartDate, CPPendDate], row_df=df_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.

    #  Fails rows 0, 1, and 3
    child_protection_plans = pd.DataFrame(
        {
            "LAchildID": ["child1", "child2", "child3", "child4", "child5"],
            "CPPstartDate": [
                "08/10/1989",
                "05/12/1993",
                "05/12/1993",
                "05/12/1997",
                pd.NA,
            ],
            "CPPendDate": [
                "08/10/1989",
                "05/12/1993",
                "12/09/2022",
                "05/12/1997",
                pd.NA,
            ],
        }
    )
    child_protection_plans["CPPstartDate"] = pd.to_datetime(
        child_protection_plans["CPPstartDate"], format="%d/%m/%Y", errors="coerce"
    )
    child_protection_plans["CPPendDate"] = pd.to_datetime(
        child_protection_plans["CPPendDate"], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {ChildProtectionPlans: child_protection_plans})

    # The result contains a NamedTuple of issues encountered
    issues = result.type1_issues

    # get table name and check it. Replace ChildProtectionPlans with the name of your table.
    issue_table = issues.table
    assert issue_table == ChildProtectionPlans

    # check that the right columns were returned. Replace CPPstartDate and CPPendDate with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [CPPstartDate, CPPendDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df

    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 3
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    pd.to_datetime("08/10/1989", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("08/10/1989", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child2",
                    pd.to_datetime("05/12/1993", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("05/12/1993", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child4",
                    pd.to_datetime("05/12/1997", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("05/12/1997", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [3],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8840' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8840"
    assert (
        result.definition.message
        == "Child Protection Plan cannot start and end on the same day"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Assessments = CINTable.Assessments
AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate
LAchildID = Assessments.LAchildID
CINdetailsID = Assessments.CINdetailsID


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of '8896'
    code="8896",
    # replace Assessments with the value in the module column of the excel sheet corresponding to this rule .
    module=CINTable.Assessments,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Within one CINDetails group there are 2 or more open Assessments groups",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[AssessmentAuthorisationDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    # Replace Assessments with the name of the table you need.
    df = data_container[Assessments]
    # Before you begin, rename the index and make it a column, so that the initial row positions can be kept intact.
    df.index.name = "ROW_ID"
    df.reset_index(inplace=True)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # Within one <CINdetails> group, there must not be more than one <Assessments> group that has no <AssessmentAuthorisationDate> (N00160) recorded

    # DF_CHECK: APPLY GROUPBYs IN A SEPARATE DATAFRAME SO THAT OTHER COLUMNS ARE NOT LOST OR CORRUPTED. THEN, MAP THE RESULTS TO THE INITIAL DATAFRAME.
    df_check = df.copy()
    # get all the locations where AssessmentAuthorisationDate is null
    df_check = df_check[df_check[AssessmentAuthorisationDate].isna()]
    # We'll have to count the number of nan values per group. NaNs cannot be counted so replace them with something that can.
    # Do this only if your rule requires that you interact with a column made up of all NaNs.
    df_check[AssessmentAuthorisationDate].fillna(1, inplace=True)
    # count how many occurences of missing AssessmentAuthorisationDate per CINdetails group in each child.
    df_check = (
        df_check.groupby([LAchildID, CINdetailsID])[AssessmentAuthorisationDate]
        .count()
        .reset_index()
    )

    # when you groupby as shown above a series is returned where the columns in the round brackets become the index and the groupby result are the values.
    # resetting the index pushes the columns in the () back as columns of the dataframe and assigns the groupby result to the column in the square bracket.

    # filter out the instances where AssessmentAuthorisationDate is missing more than once in a CINdetails group.
    df_check = df_check[df_check[AssessmentAuthorisationDate] > 1]
    issue_ids = tuple(zip(df_check[LAchildID], df_check[CINdetailsID]))

    # DF_ISSUES: GET ALL THE DATA ABOUT THE LOCATIONS THAT WERE IDENTIFIED IN DF_CHECK
    df["ERROR_ID"] = tuple(zip(df[LAchildID], df[CINdetailsID]))
    df_issues = df[df["ERROR_ID"].isin(issue_ids)]

    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    # Ensure that you do not change the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_3(
        table=Assessments, columns=[AssessmentAuthorisationDate], row_df=df_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_assessments = pd.DataFrame(
        [  # child1
            {  # fail
                LAchildID: "child1",
                CINdetailsID: "cinID1",
                AssessmentAuthorisationDate: pd.NA,  # 0 first nan date in group
            },
            {  # fail
                LAchildID: "child1",
                CINdetailsID: "cinID1",
                AssessmentAuthorisationDate: pd.NA,  # 1 second nan date in group
            },
            {  # won't be flagged because there is not more than one nan authorisation date in this group.
                LAchildID: "child1",
                CINdetailsID: "cinID2",
                AssessmentAuthorisationDate: pd.NA,  # 2
            },
            # child2
            {
                LAchildID: "child2",
                CINdetailsID: "cinID1",
                AssessmentAuthorisationDate: "26/05/2021",  # 3 ignored. not nan
            },
            {  # fail
                LAchildID: "child2",
                CINdetailsID: "cinID2",
                AssessmentAuthorisationDate: pd.NA,  # 4 first nan date in group
            },
            {  # fail
                LAchildID: "child2",
                CINdetailsID: "cinID2",
                AssessmentAuthorisationDate: pd.NA,  # 5 second nan date in group
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_assessments[AssessmentAuthorisationDate] = pd.to_datetime(
        sample_assessments[AssessmentAuthorisationDate],
        format="%d/%m/%Y",
        errors="coerce",
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {Assessments: sample_assessments})

    # Use .type3_issues to check for the result of .push_type3_issues() which you used above.
    issues_list = result.type3_issues
    # Issues list contains the objects pushed in their respective order. Since push_type3 was only used once, there will be one object in issues_list.
    assert len(issues_list) == 1

    issues = issues_list[0]

    # get table name and check it. Replace ChildProtectionPlans with the name of your table.
    issue_table = issues.table
    assert issue_table == Assessments

    # check that the right columns were pushed.
    issue_columns = issues.columns
    assert issue_columns == [AssessmentAuthorisationDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "cinID1",
                ),
                "ROW_ID": [0, 1],
            },
            {
                "ERROR_ID": (
                    "child2",
                    "cinID2",
                ),
                "ROW_ID": [4, 5],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 8925 with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8896"
    assert (
        result.definition.message
        == "Within one CINDetails group there are 2 or more open Assessments groups"
    )
